@@ -4,19 +4,13 @@
 
 # A thin OO wrapper for the milter module
 
+__version__ = '0.9.2'
+
 import os
 import milter
 import thread
 
-from milter import ACCEPT,CONTINUE,REJECT,DISCARD,TEMPFAIL,	\
-  set_flags, setdbg, setbacklog, settimeout, error,	\
-  ADDHDRS, CHGBODY, ADDRCPT, DELRCPT, CHGHDRS,	\
-  V1_ACTS, V2_ACTS, CURR_ACTS
-
-try: from milter import QUARANTINE
-except: pass
-
-__version__ = '0.8.5'
+from milter import *
 
 _seq_lock = thread.allocate_lock()
 _seq = 0
@@ -29,31 +23,182 @@ def uniqueID():
   seqno = _seq = _seq + 1
   _seq_lock.release()
   return seqno
-  
-class Milter:
-  """A simple class interface to the milter module.
-  """
+
+OPTIONAL_CALLBACKS = {
+  'connect':(P_NR_CONN,P_NOCONNECT),
+  'hello':(P_NR_HELO,P_NOHELO),
+  'envfrom':(P_NR_MAIL,P_NOMAIL),
+  'envrcpt':(P_NR_RCPT,P_NORCPT),
+  'data':(P_NR_DATA,P_NODATA),
+  'unknown':(P_NR_UNKN,P_NOUNKNOWN),
+  'eoh':(P_NR_EOH,P_NOEOH),
+  'body':(P_NR_BODY,P_NOBODY),
+  'header':(P_NR_HDR,P_NOHDRS)
+}
+
+def nocallback(func):
+  try:
+    func.milter_protocol = OPTIONAL_CALLBACKS[func.__name__][1]
+  except KeyError:
+    raise ValueError(
+      '@nocallback applied to non-optional method: '+func.__name__)
+  return func
+
+def noreply(func):
+  try:
+    nr_mask = OPTIONAL_CALLBACKS[func.__name__][0]
+  except KeyErro:
+    raise ValueError(
+      '@noreply applied to non-optional method: '+func.__name__)
+  def wrapper(self,*args):
+    rc = func(self,*args)
+    if self._protocol & nr_mask: return NOREPLY
+    return rc
+  wrapper.milter_protocol = nr_mask
+  return wrapper
+
+class DisabledAction(RuntimeError):
+  pass
+
+# A do nothing Milter base class from which python milters should derive
+# unless they are using the milter C module directly.
+
+class Base(object):
+  "The core class interface to the milter module."
+
   def _setctx(self,ctx):
-    self.__ctx = ctx
+    self._ctx = ctx
+    self._actions = CURR_ACTS         # all actions enabled by default
+    self._protocol = 0                # no protocol options by default
     if ctx:
       ctx.setpriv(self)
+  def log(self,*msg): pass
+  @nocallback
+  def connect(self,hostname,family,hostaddr): return CONTINUE
+  @nocallback
+  def hello(self,hostname): return CONTINUE
+  @nocallback
+  def envfrom(self,f,*str): return CONTINUE
+  @nocallback
+  def envrcpt(self,to,*str): return CONTINUE
+  @nocallback
+  def data(self): return CONTINUE
+  @nocallback
+  def header(self,field,value): return CONTINUE
+  @nocallback
+  def eoh(self): return CONTINUE
+  @nocallback
+  def body(self,unused): return CONTINUE
+  @nocallback
+  def unknown(self,cmd): return CONTINUE
+  def eom(self): return CONTINUE
+  def abort(self): return CONTINUE
+  def close(self): return CONTINUE
+
+  # Return mask of SMFIP_N.. protocol option bits to clear for this class
+  @classmethod
+  def protocol_mask(klass):
+    try:
+      return klass._protocol_mask
+    except AttributeError:
+      p = 0
+      for func,(nr,nc) in OPTIONAL_CALLBACKS.items():
+        func = getattr(klass,func)
+        ca = getattr(func,'milter_protocol',0)
+        #print func,hex(nr),hex(nc),hex(ca)
+        p |= (nr|nc) & ~ca
+      klass._protocol_mask = p
+      return p
+    
+  # Default negotiation sets P_NO* and P_NR* for callbacks
+  # marked @nocallback and @noreply respectively
+  def negotiate(self,opts):
+    try:
+      self._actions,p,f1,f2 = opts
+      opts[1] = self._protocol = \
+              p & ~self.protocol_mask() & ~P_RCPT_REJ & ~P_HDR_LEADSPC
+      opts[2] = 0
+      opts[3] = 0
+      #self.log("Negotiated:",opts)
+    except:
+      # don't change anything if something went wrong
+      return ALL_OPTS 
+    return CONTINUE
+
+  # Milter methods which can be invoked from most callbacks
+  def getsymval(self,sym):
+    return self._ctx.getsymval(sym)
+
+  # If sendmail does not support setmlreply, then only the
+  # first msg line is used.
+  def setreply(self,rcode,xcode=None,msg=None,*ml):
+    return self._ctx.setreply(rcode,xcode,msg,*ml)
+
+  # may only be called from negotiate callback
+  def setsmlist(self,stage,macros):
+    if not self._actions & SETSMLIST: raise DisabledAction("SETSMLIST")
+    if type(macros) in (list,tuple):
+      macros = ' '.join(macros)
+    return self._ctx.setsmlist(stage,macros)
+
+  # Milter methods which can only be called from eom callback.
+  def addheader(self,field,value,idx=-1):
+    if not self._actions & ADDHDRS: raise DisabledAction("ADDHDRS")
+    return self._ctx.addheader(field,value,idx)
+
+  def chgheader(self,field,idx,value):
+    if not self._actions & CHGHDRS: raise DisabledAction("CHGHDRS")
+    return self._ctx.chgheader(field,idx,value)
+
+  def addrcpt(self,rcpt,params=None):
+    if not self._actions & ADDRCPT: raise DisabledAction("ADDRCPT")
+    return self._ctx.addrcpt(rcpt,params)
+
+  def delrcpt(self,rcpt):
+    if not self._actions & DELRCPT: raise DisabledAction("DELRCPT")
+    return self._ctx.delrcpt(rcpt)
+
+  def replacebody(self,body):
+    if not self._actions & MODBODY: raise DisabledAction("MODBODY")
+    return self._ctx.replacebody(body)
+
+  def chgfrom(self,sender,params=None):
+    if not self._actions & CHGFROM: raise DisabledAction("CHGFROM")
+    return self._ctx.chgfrom(sender,params)
+
+  # When quarantined, a message goes into the mailq as if to be delivered,
+  # but delivery is deferred until the message is unquarantined.
+  def quarantine(self,reason):
+    if not self._actions & QUARANTINE: raise DisabledAction("QUARANTINE")
+    return self._ctx.quarantine(reason)
+
+  def progress(self):
+    return self._ctx.progress()
+  
+# A logging but otherwise do nothing Milter base class included
+# for compatibility with previous versions of pymilter.
+
+class Milter(Base):
+  "A simple class interface to the milter module."
 
-  # user replaceable callbacks
   def log(self,*msg):
     print 'Milter:',
     for i in msg: print i,
     print
 
+  @noreply
   def connect(self,hostname,family,hostaddr):
     "Called for each connection to sendmail."
     self.log("connect from %s at %s" % (hostname,hostaddr))
     return CONTINUE
 
+  @noreply
   def hello(self,hostname):
     "Called after the HELO command."
     self.log("hello from %s" % hostname)
     return CONTINUE
 
+  @noreply
   def envfrom(self,f,*str):
     """Called to begin each message.
     f -> string		message sender
@@ -62,25 +207,24 @@ class Milter:
     self.log("mail from",f,str)
     return CONTINUE
 
+  @noreply
   def envrcpt(self,to,*str):
     "Called for each message recipient."
     self.log("rcpt to",to,str)
     return CONTINUE
 
+  @noreply
   def header(self,field,value):
     "Called for each message header."
     self.log("%s: %s" % (field,value))
     return CONTINUE
 
+  @noreply
   def eoh(self):
     "Called after all headers are processed."
     self.log("eoh")
     return CONTINUE
 
-  def body(self,unused):
-    "Called to transfer the message body."
-    return CONTINUE
-
   def eom(self):
     "Called at the end of message."
     self.log("eom")
@@ -96,50 +240,23 @@ class Milter:
     self.log("close")
     return CONTINUE
 
-  # Milter methods which can be invoked from callbacks
-  def getsymval(self,sym):
-    return self.__ctx.getsymval(sym)
-
-  # If sendmail does not support setmlreply, then only the
-  # first msg line is used.
-  def setreply(self,rcode,xcode=None,msg=None,*ml):
-    return self.__ctx.setreply(rcode,xcode,msg,*ml)
-
-  # Milter methods which can only be called from eom callback.
-  def addheader(self,field,value,idx=-1):
-    return self.__ctx.addheader(field,value,idx)
-
-  def chgheader(self,field,idx,value):
-    return self.__ctx.chgheader(field,idx,value)
-
-  def addrcpt(self,rcpt,params=None):
-    return self.__ctx.addrcpt(rcpt,params)
-
-  def delrcpt(self,rcpt):
-    return self.__ctx.delrcpt(rcpt)
-
-  def replacebody(self,body):
-    return self.__ctx.replacebody(body)
-
-  def chgfrom(self,sender,params=None):
-    return self.__ctx.chgfrom(sender,params)
-
-  # When quarantined, a message goes into the mailq as if to be delivered,
-  # but delivery is deferred until the message is unquarantined.
-  def quarantine(self,reason):
-    return self.__ctx.quarantine(reason)
-
-  def progress(self):
-    return self.__ctx.progress()
-
 factory = Milter
 
-def connectcallback(ctx,hostname,family,hostaddr):
+def negotiate_callback(ctx,opts):
   m = factory()
   m._setctx(ctx)
+  return m.negotiate(opts)
+
+def connect_callback(ctx,hostname,family,hostaddr,nr_mask=P_NR_CONN):
+  m = ctx.getpriv()
+  if not m:     
+    # If not already created (because the current MTA doesn't support
+    # xmfi_negotiate), create the connection object.
+    m = factory()
+    m._setctx(ctx)
   return m.connect(hostname,family,hostaddr)
 
-def closecallback(ctx):
+def close_callback(ctx):
   m = ctx.getpriv()
   if not m: return CONTINUE
   try:
@@ -196,7 +313,7 @@ def runmilter(name,socketname,timeout = 0):
 
   # The default flags set include everything
   # milter.set_flags(milter.ADDHDRS)
-  milter.set_connect_callback(connectcallback)
+  milter.set_connect_callback(connect_callback)
   milter.set_helo_callback(lambda ctx, host: ctx.getpriv().hello(host))
   # For envfrom and envrcpt, we would like to convert ESMTP parms to keyword
   # parms, but then all existing users would have to include **kw to accept
@@ -209,12 +326,16 @@ def runmilter(name,socketname,timeout = 0):
   milter.set_body_callback(lambda ctx,chunk: ctx.getpriv().body(chunk))
   milter.set_eom_callback(lambda ctx: ctx.getpriv().eom())
   milter.set_abort_callback(lambda ctx: ctx.getpriv().abort())
-  milter.set_close_callback(closecallback)
+  milter.set_close_callback(close_callback)
 
   milter.setconn(socketname)
   if timeout > 0: milter.settimeout(timeout)
   # The name *must* match the X line in sendmail.cf (supposedly)
-  milter.register(name)
+  milter.register(name,
+        data=lambda ctx: ctx.getpriv().data(),
+        unknown=lambda ctx,cmd: ctx.getpriv().unknown(cmd),
+        negotiate=negotiate_callback
+  )
   start_seq = _seq
   try:
     milter.main()
